// Copyright (C) 2007-2008  Gelenberg Andrej
// 
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <stdio.h>
#include <dbus/dbus.h>
#include <string.h>
#include <unistd.h>
#include <libintl.h>

#include "util.h"
#include "nomem.h"
#include "bus.h"
#include "rule.h"

inline static enum error dbus_connect(struct err_s *err, DBusConnection **ret_conn, DBusBusType bus_type)
{
	DBusConnection *connection;
	DBusError error;

	dbus_error_init(&error);
	connection = dbus_bus_get(bus_type, &error);
	
	if ( connection == NULL )
	{
		if ( asprintf(&err->msg, "%s: %s", gettext("Error by getting bus"), error.message) == -1 )
		{
			nomem();
		}
		err->to_free = 1;
		dbus_error_free(&error);
		return ERR_BUS_CONNECTION;
	}

	*ret_conn = connection;

	return NO_ERROR;
}

inline static enum error add_rules(struct err_s *err, DBusConnection *connection, struct df_rule_el **r_el)
{
	struct df_rule_el *el = *r_el;
	struct df_rule_el *prev;
	DBusError error;
	dbus_error_init(&error);
	while ( el )
	{
		prev = el;
		el = el->next;
		dbus_bus_add_match(connection, prev->rule, &error);
		if ( dbus_error_is_set(&error) )
		{
			if ( asprintf(&err->msg,
				"%s: \"%s\": %s",
				gettext("Error by addin match rule"),
				prev->rule,
				error.message) == -1 )
			{
				nomem();
			}
			err->to_free = 1;
			dbus_error_free(&error);
			free(prev->rule);
			free(prev);
			*r_el = el;
			return ERR_BUS_ADD_RULES;
		}
		free(prev->rule);
		free(prev);
	}
	*r_el = el;
	return NO_ERROR;
}

enum error start_dbus_monitor(struct err_s *err, struct dbus_init_data *init_data)
{
	DBusConnection *connection;
	enum error retval;
	if ( (retval = dbus_connect(err, &connection, init_data->bus)) )
	{
		return retval;
	}
	if ( (retval = add_rules(err, connection, &init_data->df_rule_start)) )
	{
		dbus_connection_unref(connection);
		return retval;
	}

	if ( !dbus_connection_add_filter(connection,
			(DBusHandleMessageFunction)match_rule_filter,
			(void *)init_data->mr_root,
			NULL) )
	{
		nomem();
	}

 	 while (dbus_connection_read_write_dispatch (connection, -1))
	      ;

	dbus_connection_unref(connection);
	return NO_ERROR;
}
