// Copyright (C) 2007-2008  Gelenberg Andrej
// 
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

#ifndef MATCH_H
#define MATCH_H

#include <dbus/dbus.h>
#include <sys/types.h>
#include <regex.h>

enum arg_type {
	STRING = DBUS_TYPE_STRING,
	REGEX,
	BYTE = DBUS_TYPE_BYTE,
	BOOL = DBUS_TYPE_BOOLEAN,
	INT16 = DBUS_TYPE_INT16,
	INT32 = DBUS_TYPE_INT32,
	INT64 = DBUS_TYPE_INT64,
	UINT16 = DBUS_TYPE_UINT16,
	UINT32 = DBUS_TYPE_UINT32,
	UINT64 = DBUS_TYPE_UINT64,
	DOUBLE = DBUS_TYPE_DOUBLE,
};

struct arg_data {
	enum arg_type type;
	union {
		void *ptr;
		char *str;
		char bool_d; // dbus_bool_t ist 4 byte gross, so gross muss es nicht sein
		char byte_d;
		dbus_int16_t int16_d;
		dbus_int32_t int32_d;
		dbus_int64_t int64_d;
		dbus_uint16_t uint16_d;
		dbus_uint32_t uint32_d;
		dbus_uint64_t uint64_d;
		double double_d;
		regex_t *regex;
	} data;
};

struct arg_match { 
	unsigned char num;
	struct arg_data *data;
};

enum message_type_e {
	METHOD_CALL = DBUS_MESSAGE_TYPE_METHOD_CALL, // Methodaufruf
	METHOD_RETURN = DBUS_MESSAGE_TYPE_METHOD_RETURN, // Rückgabe des Methodes
	ERROR = DBUS_MESSAGE_TYPE_ERROR, // Fehler
	SIGNAL = DBUS_MESSAGE_TYPE_SIGNAL, // Signal
	INVALID = DBUS_MESSAGE_TYPE_INVALID // Signal
}; // Typ des Nachrichtes

enum match_rule_type {
	NONE,
	PATH = DBUS_HEADER_FIELD_PATH,
	INTERFACE = DBUS_HEADER_FIELD_INTERFACE,
	MEMBER = DBUS_HEADER_FIELD_MEMBER,
	DESTINATION = DBUS_HEADER_FIELD_DESTINATION,
	SENDER = DBUS_HEADER_FIELD_SENDER,
	MESSAGE_TYPE,
	ARG,
	EXEC, // Fiktives match, das dazu da ist, um ein Komando ausführen
	FORK, // Wie EXEC, aber stat system()-Aufrufs wird fork() + exec() benutzt
	SEND  // Message zum verschicken durch DBUS
};

struct send_t {
	DBusBusType bus;
	DBusMessage *msg; // DBUS Message
};

struct match_rule {	// Matchrule, das intern benutzt wird, um zu finden,
			// welche action man ausführen muss
	enum match_rule_type type; // Typ des Matches 
	// Wenn man es rekursiv löst, kann man dies hier sparen
	// struct match_rule *prev; // vorhergehendes Match
	struct match_rule *fork; // verzweigtes Match
	struct match_rule *next; // nechstes Match
	union {
		const char *interface; // Interface, von dem Nachricht kamm
		const char *sender; // Absender des Nachrichtes
		const char *path; // Path des Nachrichtes
		const char *member; // Name des Methods oder des Signales
		const char *destination; // Quelle des Nachrichtes
		const char *cmd; // Komando zum ausführen
		struct arg_match *arg; // Argument, das mitgegeben ist.
		void *ptr; // Noch einen Nahmen für Links, die Freigegeben müssen
		struct send_t *send;
		enum message_type_e message_type;
	} data; // Daten zum Vergleichen
};


void print_rule(struct match_rule *root);
void free_match_tree(struct match_rule *root);
void print_match_rules(struct match_rule *root, int shift);
DBusHandlerResult match_rule_filter(DBusConnection *connection, DBusMessage *msg, struct match_rule *mr_root);

#endif // MATCH_H
