// Copyright (C) 2007-2008  Gelenberg Andrej
// 
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <dbus/dbus.h>
#include <expat.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <stdio.h>
#include <libintl.h>
#include <stdlib.h>
#include <limits.h>
#include <errno.h>

#include "rule.h"
#include "util.h"
#include "match.h"
#include "nomem.h"

struct df_el {	// Element, der für zusammensetzung von
			// Matchrule für dbus_bus_add_match
	unsigned short int size; // Die Länge der Zeichenfolge, die diesem Element entspricht
	struct df_el *prev; // Vorhergehendes Element
};

struct mt_stack_el {
	struct match_rule *node;
	struct mt_stack_el *prev;
};

struct userdata_st { // Userdata für Parser
	struct {
		char *buf; // Zeigt auf eine Errormessage
		char to_free; // Zeigt, ob man buf freigeben muss
	} error; // Struktur zum übergeben von Fehlermeldungen

	XML_Parser parser; // Parser, damit man ihn bei Fehlern stoppen kann

	struct {
		char get; // Ob man jezt CDATA erwartet
		size_t size; // Wie viel Daten sind schon erhalten
		char *buf; // Daten
	} cdata; // Struktur Zum CDATA-Parsen

	char empty_tag; // Zeigt, ob Tag leer sein muss

	// Folgende Variablen sind zur Zusammensetzung von Matchrules für dbus_bus_add_match
	struct df_el *cur; // letztes Matchregel
	size_t rb_s; // die Grösse von Matchrule
	char *rb_i; // Die aktuelle Position in Buffer
	char rule_buf[DBUS_MAXIMUM_MATCH_RULE_LENGTH + 1]; // Buffer für Matchrule
	struct df_rule_el *df_rules; // liste der Regeln, die ins dbus eingefügt werde müssen

	// Interne Match-Baum
	struct match_rule *root_match;
	struct match_rule *cur_match;
	struct match_rule **next_match_to_set;
	struct mt_stack_el *stack;
	char fork;

	// andere daten, die aus dem parsing gewohnwn wurden
	DBusBusType dbus_bus;
};

// ein Paar Difinitionen für internen Gebrauch
static const char dbused_str[] = "dbused";
static const char exec_str[] = "exec";
static const char send_str[] = "send";
static const char fork_str[] = "fork";
static const char match_str[] = "match";
static const char type_str[] = "type";
static const char sender_str[] = "sender";
static const char interface_str[] = "interface";
static const char member_str[] = "member";
static const char path_str[] = "path";
static const char destination_str[] = "destination";
static const char arg_str[] = "arg";
static const char bus_str[] = "bus";

static const char str_str[] = "str:";
static const char regex_str[] = "regex:";
static const char char_str[] = "char:";
static const char bool_str[] = "bool:";
static const char true_str[] = "true";
static const char false_str[] = "false";
static const char double_str[] = "double:";
static const char int_str[] = "int";
static const char byte_str[] = "byte";

static const char signal_str[] = "signal";
static const char method_call_str[] = "method_call";
static const char method_return_str[] = "method_return";
static const char error_str[] = "error";

static const char system_str[] = "system";
static const char session_str[] = "session";


inline static void set_dbus_filter(struct userdata_st *ud)
{
	// wenn da üerhaupt was zum einfügen gibt, wird weiter gemacht
	if ( ud->rb_s )
	{
		struct df_rule_el *el = ud->df_rules;
		struct df_rule_el *pel = NULL;
		const char *buf = ud->rule_buf;
		size_t buf_s = ud->rb_s;
		while(el)
		{
			// wenn nue Regel länger oder gleich schon eingefügten,
			// dann kann es schon von Eingefügten abgedeckt sein
			if ( buf_s >= el->size )
			{
				if ( ! strncmp(el->rule, buf, el->size) )
				{
					return;
				}
			}
			// wenn neue Regel kürzer ist, dann kann die schon eingefügtes
			// Regel abdecken
			else
			{
				if ( ! strncmp(buf, el->rule, buf_s) )
				{
					free(el->rule);
					el->rule = get_str_copy(buf);
					el->size = buf_s;
					return;
				}
			}
			pel = el;
			el = el->next;
		}
		el = malloc(sizeof(struct df_rule_el));
		if ( ! el )
		{
			nomem();
		}
		el->next = NULL;
		el->rule = get_str_copy(buf);
		el->size = buf_s;
		if ( pel )
		{
			pel->next = el;
		}
		else
		{
			ud->df_rules = el;
		}
	}
}

inline static void stack_push(struct userdata_st *ud)
{
	struct mt_stack_el *new_stack_el = (struct mt_stack_el*)malloc(sizeof(struct mt_stack_el));
	if ( ! new_stack_el )
	{
		nomem();
	}
	new_stack_el->node = ud->cur_match;
	new_stack_el->prev = ud->stack;
	ud->stack = new_stack_el;
}

inline static struct match_rule *gnmr_base(struct userdata_st *ud)
{
	struct match_rule *retval = (struct match_rule*)malloc(sizeof(struct match_rule));
	if ( ! retval )
	{
		nomem();
	}
	retval->next = NULL;
	retval->fork = NULL;
	retval->type = NONE;
	*(ud->next_match_to_set) = retval; // Setze nechstes Nachfolger (next oder fork)
					// für das letzte Element
	ud->cur_match = retval;
	return retval;
}

// initialisierung einer neuen match rule
inline static struct match_rule *get_new_match_rule(struct userdata_st *ud)
{
	struct match_rule *retval = gnmr_base(ud);
	ud->next_match_to_set = &retval->next; // Wohin nächstes Element kommt
	return retval;
}

inline static struct match_rule *add_action(struct userdata_st *ud)
{
	set_dbus_filter(ud);
	struct match_rule *mrule = gnmr_base(ud);
	ud->next_match_to_set = &mrule->fork;
	if ( ud->fork )
	{
		ud->stack->node = mrule;
	}
	else
	{
		stack_push(ud);
		ud->fork = 1;
	}
	return mrule;
}

// Fügt einen Filter schon auf ebene von dem DBUS
static size_t add_match(struct userdata_st *ud, const char *name, const char *attr, size_t s)
{
	size_t attr_s = strlen(attr); // die Länge von Attribute
	size_t rule_s = 3 + s + attr_s; // die Länge von dem ganzem Teil des Matchrule,
			// das dazu eingefügt werden soll
			// (einschießlich Apostrophe und Gleichheitszeichen)
	size_t rb_s = ud->rb_s; // Hole eine Kopie von der Größe des Matchrul bis jetzt
	if ( rb_s != 0 )
	{
		++rule_s; // Man muss auch Komma zwischen Regeln miteinkallkulieren
			// Ntürlich nur dann, wenn es nicht die erste Regel.
	}
	// Die Länge von Matchrule ist begrenzt (die Gränze ist in header-Dateie von DBUS definiert)
	if ( ( rb_s + rule_s ) <=  DBUS_MAXIMUM_MATCH_RULE_LENGTH )
	{
		char *i = ud->rb_i; // Gegenwärtige Position in Matchrulbuffer
		if ( rb_s != 0 )
		{
			// wenn es nicht die erste Regel ist,
			// dann muss ein Komma vor der neuen Regel
			*i++ = ',';
		}
		memcpy(i, name, s); // Kopieren die Name
		i += s;
		*i = '=';
		*++i = '\'';
		memcpy(++i, attr, attr_s); // Kopiere Wert
		i += attr_s;
		*i = '\'';
		*++i = '\0';
		ud->rb_i = i;
		ud->rb_s += rule_s; // Neue Größe der Regel
		return rule_s; // Um die Größe der ganzen Regeln, die eingefügt wurden,
			// des match-tags zu berechnen
			// (wird später zur Zurücksetzung von Bufferzeiger gebraucht)
	}
	// Wenn die Gränze überschritten wird, dann wird dieser Regel so behandelt,
	// als hätte es die Länge 0
	return 0;
}

static struct arg_data *get_arg_data(struct userdata_st *ud, const char * const arg)
{
	const char *a = arg;
	struct arg_data *retval = malloc(sizeof(struct arg_data));
	// Stringargument
	if ( ! strncmp (&str_str[0], a, sizeof(str_str) - 1) )
	{
		a += sizeof(str_str) - 1;
		retval->type = STRING;
		retval->data.str = get_str_copy(a);
		if ( verbose )
		{
			printf("*\t%s: %s\n", gettext("add string match"), a);
		}
	}
	else if ( ! strncmp(&regex_str[0], a, sizeof(regex_str) - 1) )
	{
		a += sizeof(regex_str) - 1;
		retval->type = REGEX;
		regex_t *re = malloc(sizeof(regex_t));
		if ( ! re )
		{
			nomem();
		}
		int err;
		if ( (err = regcomp(re, a, REG_EXTENDED|REG_NOSUB)) )
		{
			free(retval);
			size_t s = regerror(err, re, 0, 0);
			char *err_buf = malloc(s+1);
			if ( ! err_buf )
			{
				nomem();
			}
			s = regerror(err, re, err_buf, s+1);
			if ( asprintf(&ud->error.buf, "%s: %s", gettext("Regex error"), err_buf) == -1 )
			{
				nomem();
			}
			ud->error.to_free = 1;
			XML_StopParser(ud->parser, XML_FALSE);
			free(err_buf);
			free(re);
			return NULL;
		}
		retval->data.regex = re;
		if ( verbose )
		{
			printf("*\t%s: %s\n", gettext("add regex arg match"), a);
		}
	}
	else if ( ! strncmp(&char_str[0], a, sizeof(char_str) -1) )
	{
		a += sizeof(char_str) - 1;
		retval->type = BYTE;
		retval->data.byte_d = *a++;
		if ( *a )
		{
			free(retval);
			ud->error.buf = gettext("No single char was found");
			ud->error.to_free = 0;
			XML_StopParser(ud->parser, XML_FALSE);
			return NULL;
		}
		if ( verbose )
		{
			printf("*\t%s: %c\n", gettext("add char match"), *a);
		}
	}
	else if ( ! strncmp(&bool_str[0], a, sizeof(bool_str) -1) )
	{
		a += sizeof(bool_str) - 1;
		retval->type = BOOL;
		if ( !strncmp(&true_str[0], a, sizeof(true_str)) )
		{
			retval->data.bool_d = 1;
		}
		else if ( !strncmp(&false_str[0], a, sizeof(false_str)) )
		{
			retval->data.bool_d = 0;
		}
		else
		{
			free(retval);
			ud->error.buf = gettext("No valid boolean value found");
			ud->error.to_free = 0;
			XML_StopParser(ud->parser, XML_FALSE);
			return NULL;
		}
		if ( verbose )
		{
			printf("*\t%s: %s\n", gettext("add boolean match"), a);
		}

	}
	else if ( !strncmp(&double_str[0], a, sizeof(double_str) -1))
	{
		a += sizeof(double_str) -1;
		retval->type = DOUBLE;
		char *e;
		retval->data.double_d = strtod(a, &e);
		if ( ( e == a ) || ( *e != '\0' ) )
		{
			free(retval);
			ud->error.buf = gettext("No floating point number found");
			ud->error.to_free = 0;
			XML_StopParser(ud->parser, XML_FALSE);
			return NULL;
		}
		if ( verbose )
		{
			printf("*\t%s: %s\n", gettext("add double match"), a);
		}
	}
	else
	{
		// Intargumen
		char uns;
		if ( *a == 'u' )
		{
			uns = 1;
			++a;
		}
		else
		{
			uns = 0;
		}

		char found = 0;
		char t = 'b';
		if ( ! strncmp(&int_str[0], a, sizeof(int_str)-1) )
		{
			a += sizeof(int_str) -1;
			t = *a++;
			if (	((t == '1') && (*a == '6')) ||
					((t == '3') && (*a == '2')) ||
					((t == '6') && (*a == '4'))
				)
			{
				found = 1;
				++a;
			}
		}
		else if (!strncmp(&byte_str[0], a, sizeof(byte_str)-1))
		{
			a += sizeof(byte_str) - 1;
			found = 1;
		}
		if ( found && (*a != ':') ) 
		{
			found = 0;
		}
		else
		{
			++a;
		}

		if ( found )
		{
			char *e;
			errno = 0;
			unsigned long int val;
			val = strtoul(a, &e, 0);
			if ( ( e == a ) || ( *e != '\0' ) )
			{
				free(retval);
				ud->error.buf = gettext("No integer found");
				ud->error.to_free = 0;
				XML_StopParser(ud->parser, XML_FALSE);
				return NULL;
			}
			char oor = 0;
			switch ( t )
			{
				case 'b':
					if ( ((long int)val > CHAR_MAX) || ((long int)val < CHAR_MIN)  )
					{
						oor = 1;
						break;
					}
					retval->type = BYTE;
					if ( verbose )
					{
						printf("*\t%s: %s\n", gettext("add byte match"), a);
					}
					break;
				case '1':
					if ( (uns && (val > USHRT_MAX )) || ((!uns) &&
						(((long int)val > SHRT_MAX) || ((long int)val < SHRT_MIN))) )
					{
						oor = 1;
						break;
					}
					if ( uns )
					{
						retval->type = UINT16;
						if ( verbose )
						{
							printf("*\t%s: %s\n", gettext("add int16 match"), a);
						}
					}
					else
					{
						retval->type = INT16;
						if ( verbose )
						{
							printf("*\t%s: %s\n", gettext("add uint16 match"), a);
						}
					}
					break;
				case '3':
					if ( (uns && (val > UINT_MAX )) || ((!uns) &&
						(((long int)val > INT_MAX) || ((long int)val < INT_MIN))) )
					{
						oor = 1;
						break;
					}
					if ( uns )
					{
						retval->type = UINT32;
						if ( verbose )
						{
							printf("*\t%s: %s\n", gettext("add int32 match"), a);
						}
					}
					else
					{
						retval->type = INT32;
						if ( verbose )
						{
							printf("*\t%s: %s\n", gettext("add uint32 match"), a);
						}
					}
					break;
				case '6':
					if ( errno == ERANGE )
					{
						oor = 1;
						break;
					}
					if ( uns )
					{
						retval->type = UINT64;
						if ( verbose )
						{
							printf("*\t%s: %s\n", gettext("add int64 match"), a);
						}
					}
					else
					{
						retval->type = INT64;
						if ( verbose )
						{
							printf("*\t%s: %s\n", gettext("add uint64 match"), a);
						}
					}
					break;
			}
			retval->data.uint64_d = val;
			if ( oor )
			{
				free(retval);
				ud->error.buf = gettext("Integer is out of range");
				ud->error.to_free = 0;
				XML_StopParser(ud->parser, XML_FALSE);
				return NULL;
			}
		}
		else
		{
			a = arg;
			// Unbekanter Argumenttyp
			while( *a++ != ':' ) // Suche nach Argumenttyp
			{
				if ( *a == 0 ) // nix gefunden
				{
					free(retval);
					ud->error.buf = gettext("Argument type not found");
					ud->error.to_free = 0;
					XML_StopParser(ud->parser, XML_FALSE);
					return NULL;
				}
			}
			size_t s = a - arg;
			char *type = (char*)malloc(s);
			memcpy(type, arg, --s);
			*(type + s) = '\0';
			if ( asprintf(	&ud->error.buf,
				"%s: %s",
				gettext("Unknown argument type"),
				type) == -1 )
			{
				nomem();
			}
			free(retval);
			free(type);
			ud->error.to_free = 1;
			XML_StopParser(ud->parser, XML_FALSE);
			return NULL;
		}
	}
	return retval;
}

static void freeSendHandler(struct arg_data **a, unsigned char n)
{
	if ( n )
	{
		struct arg_data *b;
		struct arg_data **d = a;
		int i;
		for(i = 0; i < n; ++i, ++d)
		{
			b = *d;
			if ( b )
			{
				if ( b->type == STRING )
				{
					free(b->data.str);
				}
				free(b);
			}
		}
		free(a);
	}
}

inline static void SendHandler(const char **atts, struct userdata_st *ud)
{
	enum message_type_e type = INVALID;
	struct arg_data **args = NULL;
	unsigned char args_s = 0;
	struct arg_data **arg;
	struct arg_data *new_arg;
	const char *path = NULL;
	const char *interface = NULL;
	const char *member = NULL;
	const char *destination = NULL;
	const char *sender = NULL;
	const char *bus = NULL;
	
	const char *attr_name;
	const char **i;

	ud->empty_tag = 1;
	
	for( i=atts; *i; ++i )
	{
		attr_name = *i++;
		if ( ! strcmp(&member_str[0], attr_name) )
		{
			if ( member )
			{
				freeSendHandler(args, args_s);
				ud->error.buf = gettext("Member is alredy set");
				ud->error.to_free = 0;
				return;
			}
			member = *i;
		}
		else if ( ! strcmp(&destination_str[0], attr_name) )
		{
			if ( destination )
			{
				freeSendHandler(args, args_s);
				ud->error.buf = gettext("Destination is alredy set");
				ud->error.to_free = 0;
				return;
			}
			destination = *i;
		}
		else if ( ! strcmp(&interface_str[0], attr_name) )
		{
			if ( interface )
			{
				freeSendHandler(args, args_s);
				ud->error.buf = gettext("Interface is alredy set");
				ud->error.to_free = 0;
				return;
			}
			interface = *i;
		}
		else if ( ! strcmp(&path_str[0], attr_name) )
		{
			if ( path )
			{
				freeSendHandler(args, args_s);
				ud->error.buf = gettext("Path is alredy set");
				ud->error.to_free = 0;
				return;
			}
			path = *i;
		}
		else if ( ! strcmp(&sender_str[0], attr_name) )
		{
			if ( sender )
			{
				freeSendHandler(args, args_s);
				ud->error.buf = gettext("Sender is alredy set");
				ud->error.to_free = 0;
				return;
			}
			sender = *i;
		}
		else if ( ! strcmp(&bus_str[0], attr_name) )
		{
			if ( bus )
			{
				freeSendHandler(args, args_s);
				ud->error.buf = gettext("Bus is alredy set");
				ud->error.to_free = 0;
				return;
			}
			bus = *i;
		}
		else if ( ! strcmp(&type_str[0], attr_name) )
		{
			if ( type != INVALID )
			{
				freeSendHandler(args, args_s);
				ud->error.buf = gettext("Type is alredy set");
				ud->error.to_free = 0;
				return;
			}
			const char *str = *i;
			if ( ! strcmp(&method_call_str[0], str) )
			{
				type = METHOD_CALL;
			}
			else if ( ! strcmp(&signal_str[0], str) )
			{
				type = SIGNAL;
			}
			else
			{
				freeSendHandler(args, args_s);
				if ( asprintf(&ud->error.buf,
					"%s \"%s\"",
					gettext("Unsupported message type"),
					str) == -1 )
				{
					nomem();
				}
				ud->error.to_free = 1;
				XML_StopParser(ud->parser, XML_FALSE);
				return;
			}
		}
		else if (! strncmp(&arg_str[0], attr_name, sizeof(arg_str) - 1) )
		{
			attr_name += sizeof(arg_str) - 1;
			unsigned long n = atol(attr_name);
			if ( n > DBUS_MAXIMUM_MATCH_RULE_ARG_NUMBER )
			{
				freeSendHandler(args, args_s);
				ud->error.buf = gettext("Too many argumens");
				ud->error.to_free = 0;
				XML_StopParser(ud->parser, XML_FALSE);
				return;
			}
			if ( n >= args_s )
			{
				args = realloc(args, (n + 1) * sizeof(struct arg_data));
				arg = args + args_s;
				unsigned char i;
				for(i = n + 1 - args_s; i; --i)
				{
					*arg++ = NULL;
				}
				args_s = n + 1;
			}
			arg = args + n;
			if ( *arg )
			{
				freeSendHandler(args, args_s);
				if ( asprintf(&ud->error.buf,
					"%s: %s",
					gettext("Argumen is alredy set"),
					attr_name) == -1 )
				{
					nomem();
				}
				ud->error.to_free = 1;
				XML_StopParser(ud->parser, XML_FALSE);
				return;
			}
			if ( verbose )
			{
				printf("* %s %ld\n", gettext("get argument"), n);
			}
			new_arg = get_arg_data(ud, *i);
			if ( ! new_arg )
			{
				freeSendHandler(args, args_s);
				return;
			}
			if ( new_arg->type == REGEX )
			{
				regfree(new_arg->data.regex);
				free(new_arg->data.regex);
				free(new_arg);
				freeSendHandler(args, args_s);
				ud->error.buf = gettext("Action send support not regex argument");
				ud->error.to_free = 0;
				XML_StopParser(ud->parser, XML_FALSE);
				return;
			}
			*arg = new_arg;
		}
		else
		{
			freeSendHandler(args, args_s);
			if ( asprintf(&ud->error.buf,
				"%s \"%s\"",
				gettext("Unsupported attribute"),
				attr_name) == -1 )
			{
				nomem();
			}
			ud->error.to_free = 1;
			XML_StopParser(ud->parser, XML_FALSE);
			return;
		}
	}

	if ( ! bus )
	{
		freeSendHandler(args, args_s);
		ud->error.buf = gettext("No bus is set");
		ud->error.to_free = 0;
		XML_StopParser(ud->parser, XML_FALSE);
		return;
	}

	DBusMessage *dbus_msg;
	if ( type == SIGNAL )
	{
		if ( ! path )
		{
			freeSendHandler(args, args_s);
			ud->error.buf = gettext("No path for signal");
			ud->error.to_free = 0;
			XML_StopParser(ud->parser, XML_FALSE);
			return;
		}
		if ( ! interface )
		{
			freeSendHandler(args, args_s);
			ud->error.buf = gettext("No interface for signal");
			ud->error.to_free = 0;
			XML_StopParser(ud->parser, XML_FALSE);
			return;
		}
		if ( ! member )
		{
			freeSendHandler(args, args_s);
			ud->error.buf = gettext("No member for signal");
			ud->error.to_free = 0;
			XML_StopParser(ud->parser, XML_FALSE);
			return;
		}
		dbus_msg = dbus_message_new_signal(path, interface, member);
	}
	else
	{
		if ( ! path )
		{
			freeSendHandler(args, args_s);
			ud->error.buf = gettext("No path for method call");
			ud->error.to_free = 0;
			XML_StopParser(ud->parser, XML_FALSE);
			return;
		}
		if ( ! member )
		{
			freeSendHandler(args, args_s);
			ud->error.buf = gettext("No member for method call");
			ud->error.to_free = 0;
			XML_StopParser(ud->parser, XML_FALSE);
			return;
		}
		dbus_msg = dbus_message_new_method_call(destination, path,
				interface, member);
	}
	if ( ! dbus_msg )
	{
		nomem();
	}

	DBusMessageIter iter;
	dbus_message_iter_init_append(dbus_msg, &iter);
	
	unsigned char k;
	struct arg_data **arg_i = args;
	for( k=0; k < args_s; ++k, ++arg_i)
	{
		new_arg = *arg_i;
		if ( ! new_arg )
		{
			freeSendHandler(args, args_s);
			dbus_message_unref(dbus_msg);
			ud->error.buf = gettext("Not all argument are set");
			ud->error.to_free = 0;
			XML_StopParser(ud->parser, XML_FALSE);
			return;
		}
		dbus_message_iter_append_basic(&iter, new_arg->type, &new_arg->data.ptr);
	}

	struct send_t *s = malloc(sizeof(struct send_t));
	s->msg = dbus_msg;

	if ( ! strcmp(&system_str[0], bus) )
	{
		s->bus = DBUS_BUS_SYSTEM;
	}
	else if ( ! strcmp(&session_str[0], bus) )
	{
		s->bus = DBUS_BUS_SESSION;
	}
	else
	{
		free(s);
		freeSendHandler(args, args_s);
		if ( asprintf(	&ud->error.buf,
				"%s: %s",
				gettext("Unknown dbus bus"),
				bus) == -1 )
		{
			nomem();
		}
		ud->error.to_free = 1;
		XML_StopParser(ud->parser, XML_FALSE);
		return;
	}

	struct match_rule *mr = add_action(ud);
	mr->type = SEND;
	mr->data.send = s;

	freeSendHandler(args, args_s);
}

inline static void DEDHandler(const char **atts, struct userdata_st *ud)
{
	const char **i;
	char bus_is_not_set = 1;
	for(i = atts; *i; ++i)
	{
		if ( ! strcmp(&bus_str[0], *i++) )
		{
			if ( ! strcmp(&system_str[0], *i) )
			{
				ud->dbus_bus = DBUS_BUS_SYSTEM;
			}
			else if ( ! strcmp(&session_str[0], *i) )
			{
				ud->dbus_bus = DBUS_BUS_SESSION;
			}
			else
			{
				if ( asprintf(	&ud->error.buf,
						"%s: %s",
						gettext("Unknown dbus bus"),
						*i) == -1 )
				{
					nomem();
				}
				ud->error.to_free = 1;
				XML_StopParser(ud->parser, XML_FALSE);
				return;
			}
			bus_is_not_set = 0;
			if ( verbose )
			{
				printf("* %s: %s\n", gettext("set dbus bus"), *i);
			}
		}
		else
		{
			// Es ist ein unbekanter Attribut aufgetretten
			if ( asprintf(	&ud->error.buf,
					"%s: %s",
					gettext("Unknown attribute"),
					*i++) == -1 )
			{
				nomem();
			}
			ud->error.to_free = 1;
			XML_StopParser(ud->parser, XML_FALSE);
			return;
		}
	}
	if (bus_is_not_set)
	{
		ud->error.buf = gettext("Message bus ist not set");
		ud->error.to_free = 0;
		XML_StopParser(ud->parser, XML_FALSE);
		return;
	}
}

// bearbeitet arg-Attribute von match-tag
inline static size_t argMatch_h(struct match_rule *mrule, const char *attr_name, const char *arg, struct userdata_st *ud)
{
	size_t retval = 0;
	struct arg_match *arg_m = malloc(sizeof(struct arg_match));
	if ( ! arg_m )
	{
		nomem();
	}
	unsigned long n = atol(attr_name + (sizeof(arg_str) - 1));
	if ( n > DBUS_MAXIMUM_MATCH_RULE_ARG_NUMBER )
	{
		free(arg_m);
		ud->error.buf = gettext("Too many argumens to match");
		ud->error.to_free = 0;
		XML_StopParser(ud->parser, XML_FALSE);
		return -1;
	}
	arg_m->num = n;
	if ( verbose )
	{
		printf("* %s %ld\n", gettext("get argument"), n);
	}
	struct arg_data *arg_d = get_arg_data(ud, arg);
	if ( ! arg_d )
	{
		free(arg_m);
		return -1;
	}
	if ( arg_d->type == STRING)
	{
		retval = add_match(ud, attr_name, arg_d->data.str, strlen(attr_name));
	}
	arg_m->data = arg_d;

	mrule->type = ARG;
	mrule->data.arg = arg_m;
	return retval;
}


// Bearbeitung von match-tag
inline static void StartMatch(const char **atts, struct userdata_st *ud)
{
	char fork = ud->fork;
	char to_push;

	if ( fork )
	{
		ud->fork = 0;
		to_push = 0;
	}
	else
	{
		to_push = 1;
	}

	const char *attr_name; // Variable für die Name des Attributes
	const char **i; // Laufvariable durch name-wert des Attributes
	struct match_rule *mrule;

	size_t match_s = 0; // die Länge des Matchregels für DBUS

	// gehe durch alle attribute, die Tag hat
	for (i = atts; *i; ++i)
	{
		mrule = get_new_match_rule(ud);
		attr_name = *i++;
		// member-Attribut gefunden
		if ( ! strcmp(&member_str[0], attr_name) )
		{
			mrule->type = MEMBER;
			mrule->data.member = get_str_copy(*i);
			match_s += add_match(ud, &member_str[0], *i, sizeof(member_str) - 1);
		}
		// destination-Attribut gefunden
		else if ( ! strcmp(&destination_str[0], attr_name) )
		{
			mrule->type = DESTINATION;
			mrule->data.destination = get_str_copy(*i);
			match_s += add_match(ud, &destination_str[0], *i, sizeof(destination_str) - 1);
		}
		// arg*-Attribut gefunden
		else if ( ! strncmp(&arg_str[0], attr_name, sizeof(arg_str) - 1) )
		{
			size_t res = argMatch_h(mrule, attr_name, *i, ud);
			if ( res  == -1 )
			{
				return;
			}
			match_s += res;
		}
		// interface-Attribut gefunden
		else if ( ! strcmp(&interface_str[0], attr_name) )
		{
			mrule->type = INTERFACE;
			mrule->data.interface = get_str_copy(*i);
			match_s += add_match(ud, &interface_str[0], *i, sizeof(interface_str) - 1);
		}
		// sender-Attribut gefunden
		else if ( ! strcmp(&sender_str[0], attr_name) )
		{
			mrule->type = SENDER;
			mrule->data.sender = get_str_copy(*i);
			match_s += add_match(ud, &sender_str[0], *i, sizeof(sender_str) - 1);
		}
		// patch-Attribut gefunden
		else if ( ! strcmp(&path_str[0], attr_name) )
		{
			mrule->type = PATH;
			mrule->data.path = get_str_copy(*i);
			match_s += add_match(ud, &path_str[0], *i, sizeof(path_str) - 1);
		}
		// match-Attribut gefunden
		else if ( ! strcmp(&type_str[0], attr_name) )
		{
			const char *str = *i;
			mrule->type = MESSAGE_TYPE;
			if ( ! strcmp(&method_call_str[0], str) )
			{
				mrule->data.message_type = METHOD_CALL;
			}
			else if ( ! strcmp(&method_return_str[0], str) )
			{
				mrule->data.message_type = METHOD_RETURN;
			}
			else if ( ! strcmp(&error_str[0], str) )
			{
				mrule->data.message_type = ERROR;
			}
			else if ( ! strcmp(&signal_str[0], str) )
			{
				mrule->data.message_type = SIGNAL;
			}
			else
			{
				if ( asprintf(&ud->error.buf,
					"%s \"%s\"",
					gettext("Unsupported message type"),
					str) == -1 )
				{
					nomem();
				}
				ud->error.to_free = 1;
				XML_StopParser(ud->parser, XML_FALSE);
				return;
			}
			match_s += add_match(ud, &type_str[0], *i, sizeof(type_str) - 1);
		}
		else
		{
			if ( asprintf(&ud->error.buf,
				"%s: %s",
				gettext("Unknown attribute"),
				attr_name) == -1 )
			{
				nomem();
			}
			ud->error.to_free = 1;
			XML_StopParser(ud->parser, XML_FALSE);
			return;
		}

		if ( fork )
		{
			ud->stack->node = mrule;
			fork = 0;
		}

		if ( to_push )
		{
			stack_push(ud);
			to_push = 0;
		}
	}
	struct df_el *le = malloc(sizeof(struct df_el));
	le->size = match_s; // setzte die Länge 
	le->prev = ud->cur;
	ud->cur = le;
}

static void CdataHandler(struct userdata_st *ud, const char *data, int len)
{
	if ( ud->empty_tag )
	{
		ud->error.buf = gettext("Parsingerror: element must be empty, CDATA is get");
		ud->error.to_free = 0;
		XML_StopParser(ud->parser, XML_FALSE);
		return;
	}

	if ( ud->cdata.get )
	{
		if ( ud->cdata.buf )
		{
			size_t ns = ud->cdata.size + len;
			ud->cdata.buf = realloc(ud->cdata.buf, ns + 1);
			if ( ! ud->cdata.buf )
			{
				nomem();
			}
			memcpy(ud->cdata.buf + ud->cdata.size, data, len);
			*(ud->cdata.buf + ns) = '\0';
			ud->cdata.size = ns;
		}
		else
		{
			ud->cdata.buf = malloc(len + 1);
			memcpy(ud->cdata.buf, data, len);
			*(ud->cdata.buf + len) = '\0';
			ud->cdata.size = len;
		}
	}
}

static void ElementStart(struct userdata_st *ud, const char *name, const char **atts)
{
	if ( ud->empty_tag )
	{
		ud->error.buf = gettext("Parsingerror: element must be empty");
		ud->error.to_free = 0;
		XML_StopParser(ud->parser, XML_FALSE);
		return;
	}

	if ( ud->cdata.get ) // Die Prüfung, ob es CDATA erwartet wird 
	{
		ud->error.buf = gettext("Parsingerror: only CDATA waited");
		ud->error.to_free = 0;
		if ( ud->cdata.buf )
		{
			free(ud->cdata.buf);
			ud->cdata.buf = NULL;
		}
		XML_StopParser(ud->parser, XML_FALSE);
		return;
	}
	else if ( ! strcmp(&exec_str[0], name) )
	{
		ud->cdata.get = 1;
		return;
	}
	else if ( ! strcmp(&fork_str[0], name) )
	{
		ud->cdata.get = 1;
		return;
	}
	else if ( ! strcmp(&send_str[0], name) )
	{
		SendHandler(atts, ud);
		return;
	}
	else if ( ! strcmp(&dbused_str[0], name) )
	{
		DEDHandler(atts, ud);
		return;
	}
	else if ( ! strcmp(&match_str[0], name) )
	{
		StartMatch(atts, ud);
		return;
	}
	// Es handelt sich um einem nicht bekannten Element
	if ( asprintf(&ud->error.buf,
		"%s \"%s\"",
		gettext("Unknown Element"),
		name) == -1 )
	{
		nomem();
	}
	ud->error.to_free = 1;
	XML_StopParser(ud->parser, XML_FALSE);
}

static void ElementEnd(struct userdata_st *ud, const char *name)
{
	if ( ! strcmp(&match_str[0], name) )
	{
		// start behandlung von dem Stack für Dbusfilter
		struct df_el *le = ud->cur;
		size_t s = le->size;
		ud->rb_s -= s;
		ud->rb_i -= s;
		*ud->rb_i = '\0';
		ud->cur = le->prev;
		free(le);
		// ende

		// behandlung von dem internen Filter
		if ( ud->fork )
		{
			struct mt_stack_el *el = ud->stack;
			ud->stack = el->prev;
			ud->cur_match = ud->stack->node;
			free(el);
		}
		ud->next_match_to_set = &ud->cur_match->fork;
		ud->fork = 1;
	}
	else if ( ! strcmp(&exec_str[0], name) )
	{
		struct match_rule *mr = add_action(ud);
		mr->type = EXEC;
		mr->data.cmd = get_str_copy(ud->cdata.buf);
	}
	else if ( ! strcmp(&fork_str[0], name) )
	{
		struct match_rule *mr = add_action(ud);
		mr->type = FORK;
		mr->data.cmd = get_str_copy(ud->cdata.buf);
	}
	else if ( ! strcmp(&send_str[0], name) )
	{
		ud->empty_tag = 0;
	}
	if ( ud->cdata.buf )
	{
		free(ud->cdata.buf);
		ud->cdata.buf = NULL;
	}
	ud->cdata.get = 0;
}

inline static XML_Parser parser_init()
{
	XML_Parser parser = XML_ParserCreate(NULL);
	XML_SetParamEntityParsing(parser, XML_PARAM_ENTITY_PARSING_NEVER); // Kein Parsing von Entitys
	XML_UseForeignDTD(parser, XML_FALSE); // Kein DTD-Parsing
	XML_SetStartElementHandler(parser, (XML_StartElementHandler)&ElementStart);
	XML_SetEndElementHandler(parser, (XML_EndElementHandler)&ElementEnd);
	XML_SetCharacterDataHandler(parser, (XML_CharacterDataHandler)&CdataHandler);
	return parser;
}

inline static void free_stack(struct mt_stack_el *el)
{
	struct mt_stack_el *b;
	while(el)
	{
		b = el->prev;
		free(el);
		el = b;
	}
}

inline static void free_list(struct df_rule_el *start)
{
	struct df_rule_el *prev;
	while(start)
	{
		prev = start;
		start = start->next;
		free(prev->rule);
		free(prev);
	}
}

inline static void lr_err_free(struct userdata_st *ud)
{
	free_list(ud->df_rules);
	free_stack(ud->stack);
	free_match_tree(ud->root_match);
}

enum error load_rules(const char *file_name, struct dbus_init_data *retval, struct err_s *error)
{
	int fd = open(file_name, O_RDONLY);
	if ( fd == -1 )
	{
		error->msg = strerror(errno);
		error->to_free = 0;
		return ERR_LR_OPEN;
	}

	size_t buf_size;
	struct stat st_buf;
	if ( fstat(fd, &st_buf) )
	{
		error->msg = strerror(errno);
		error->to_free = 0;
		return ERR_LR_STAT;
	}

	if ( ! S_ISREG(st_buf.st_mode) ) // Prüfung, ob es sich um ein reguleres Datei handelt
	{
		error->msg = gettext("is not regular file");
		error->to_free = 0;
		return ERR_LR_STAT;
	}
	buf_size = st_buf.st_size <= 2097152 ? st_buf.st_size : 1048576 ;
	// Wenn Datei kleiner oder gleich 2Mb gross ist, dann wird es vollständig ins Speicher geladen,
	// wenn Datei größer als 2 Mb ist, wird das Puffer auf 1 Mb gesetzt

	char *buf = malloc(buf_size);
	if ( ! buf )
	{
		buf_size = 1024;
		// Wenn es nich genug speicher für 1Mb Puffer gibt,
		// dann probiere mal mit 1 Kb
		buf = malloc(buf_size);
		if ( ! buf )
		{
			nomem();
		}
	}

	XML_Parser parser = parser_init();

	// Initialisierung von Userdata, das beim Parsing gebraucht wird
	struct userdata_st ud;
	ud.error.buf = NULL;
	ud.cdata.get = 0;
	ud.cdata.size = 0;
	ud.cdata.buf = NULL;

	ud.empty_tag = 0;

	ud.cur = NULL;
	ud.rb_s = 0;
	ud.rb_i = &ud.rule_buf[0];
	ud.df_rules = NULL;

	ud.cur_match = NULL;
	ud.root_match = NULL;
	ud.next_match_to_set = &ud.root_match;
	ud.stack = NULL;
	ud.fork = 0;
	ud.parser = parser;

	XML_SetUserData(parser, (void*)&ud);

	size_t count = 0;
	while ( count < st_buf.st_size )
	{
		size_t res = read(fd, buf, buf_size);
		if ( res == -1 )
		{
			error->msg = strerror(errno);
			error->to_free = 0;
			free(buf);
			close(fd);
			XML_ParserFree(parser);
			lr_err_free(&ud);
			return ERR_LR_READ;
		}
		count += res;
		enum XML_Status stat = XML_Parse(parser, buf, res, ( count >= st_buf.st_size ));
		if ( ! stat )
		{
			free(buf);
			close(fd);
			if ( ud.error.buf )
			{
				if ( asprintf(&error->msg, "%s %d, %s %d: %s",
					gettext("line"),
					(int)XML_GetCurrentLineNumber(parser),
					gettext("col"),
					(int)XML_GetCurrentColumnNumber(parser),
					ud.error.buf) == -1 )
				{
					nomem();
				}
				if ( ud.error.to_free)
				{
					free(ud.error.buf);
				}
			}
			else
			{
				enum XML_Error p_err = XML_GetErrorCode(parser);
				if ( asprintf(&error->msg, "%s %d, %s %d: %s",
					gettext("line"),
					(int)XML_GetCurrentLineNumber(parser),
					gettext("col"),
					(int)XML_GetCurrentColumnNumber(parser),
					XML_ErrorString(p_err)) == -1 )
				{
					nomem();
				}
			}
			error->to_free = 1;
			XML_ParserFree(parser);

			// Befreie übrig gebliebene Anträge von df_el
			struct df_el *le;
			while ( ud.cur )
			{
				le = ud.cur;
				ud.cur = le->prev;
				free(le);
			}
			lr_err_free(&ud);
			return ERR_LR_PARSE;
		}
	}

	free(ud.stack);

	retval->mr_root = ud.root_match;
	retval->df_rule_start = ud.df_rules;
	retval->bus = ud.dbus_bus;

	free(buf);
	close(fd);
	XML_ParserFree(parser);

	return NO_ERROR;
}
